#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <cls.h>
#include <hash.h>
#include <jmp.h>
#include <ui.h>
#include <method.h>
#include <arena.h>
#include <cls.h>
#include <obj.h>
#include <jthread.h>
#include <mvector.h>
#include <methodtime.h>
#include <heap_dump.h>
#include <jmp-config.h>

static char* Q = "?";
static char* SYS_INIT = "[system init]";
static char sarena[64];
static char is_array[64];
static char size[64];
static char buf[64];

static obj* last_selected_object;

enum {
    OD_CLASS,
    OD_ARENA,
    OD_IS_ARRAY,
    OD_SIZE,
    OD_ALLOC_CLASS,
    OD_ALLOC_METHOD,
    OD_VARIABLE,
    OD_OBJECT,
    ODN_COLUMNS
};

typedef struct {
    GtkWidget *window;
    GtkTreeStore *clist;
    gint       row;
    void      *data;
} wret;

typedef struct {
    GtkTreeStore *clist;
    gint       row;    
} col_click_info;

static gint sort_by_class_name (GtkTreeModel *model,
				GtkTreeIter *a,
				GtkTreeIter *b,
				gpointer user_data) {
    obj* o1;
    obj* o2;
    cls* cp1;
    cls* cp2;
    char* cn1;
    char* cn2;

    gtk_tree_model_get (model, a, OD_OBJECT, &o1, -1);
    gtk_tree_model_get (model, b, OD_OBJECT, &o2, -1); 
    cp1 = get_class (obj_get_class_id (o1));
    cp2 = get_class (obj_get_class_id (o2));
    cn1 = cp1 ? cp1->name : Q;
    cn2 = cp2 ? cp2->name : Q;
    return (strcmp (cn1, cn2));    
}


static gint sort_by_arena (GtkTreeModel *model,
			   GtkTreeIter *a,
			   GtkTreeIter *b,
			   gpointer user_data) {
    obj* o1;
    obj* o2;
    gtk_tree_model_get (model, a, OD_OBJECT, &o1, -1);
    gtk_tree_model_get (model, b, OD_OBJECT, &o2, -1); 
    return o1->arena_id - o2->arena_id;
}

static gint sort_by_is_array (GtkTreeModel *model,
			      GtkTreeIter *a,
			      GtkTreeIter *b,
			      gpointer user_data) {
    obj* o1;
    obj* o2;
    gtk_tree_model_get (model, a, OD_OBJECT, &o1, -1);
    gtk_tree_model_get (model, b, OD_OBJECT, &o2, -1); 
    return o1->is_array - o2->is_array;
}

static gint sort_by_size (GtkTreeModel *model,
			  GtkTreeIter *a,
			  GtkTreeIter *b,
			  gpointer user_data) {
    obj* o1;
    obj* o2;
    gtk_tree_model_get (model, a, OD_OBJECT, &o1, -1);
    gtk_tree_model_get (model, b, OD_OBJECT, &o2, -1); 
    return o1->size - o2->size;
}

static gint sort_by_alloc_class_and_method (GtkTreeModel *model,
					    GtkTreeIter *a,
					    GtkTreeIter *b,
					    gpointer user_data) {
    cls* c1 = NULL;
    cls* c2 = NULL;
    char* cn1;
    char* cn2;
    int i;
    obj* o1;
    obj* o2;
    method* m1;
    method* m2;
    gtk_tree_model_get (model, a, OD_OBJECT, &o1, -1);
    gtk_tree_model_get (model, b, OD_OBJECT, &o2, -1); 
    m1 = o1->method;
    m2 = o2->method;
    if (m1)
	c1 = m1->owner;
    if (m2)
	c2 = m2->owner;    
    cn1 = c1 ? c1->name : Q;
    cn2 = c2 ? c2->name : Q;
    i = strcmp (cn1, cn2);
    if (i)
	return i;
    cn1 = m1 ? m1->jmpname : SYS_INIT;
    cn2 = m2 ? m2->jmpname : SYS_INIT;
    i = strcmp (cn1, cn2);
    return i;
}

static gint ((*obj_comprs[])(GtkTreeModel *model,
			     GtkTreeIter *a,
			     GtkTreeIter *b,
			     gpointer user_data)) = { sort_by_class_name,
						      sort_by_arena, 
						      sort_by_is_array,          /* is_array */
						      sort_by_size,              /* size */
						      sort_by_alloc_class_and_method,  /* alloc_class */
						      sort_by_alloc_class_and_method,   /* alloc_method */
						      NULL                       /* variable */
			     };


static void append_object (GtkTreeStore* list, GtkTreeIter* iter, GtkTreeIter* parent, 
			   obj* o, char* variable) {
    cls* c = NULL;
    arena* a = NULL;
    cls* cp = get_class (obj_get_class_id (o));
    gchar *text[ODN_COLUMNS];

    /* titles = { "Class name", "arena", "is array", "size", "allocating class", "allocating method"};*/    
    method* m = o->method;
    if (m != NULL)
	c = m->owner;
    text[OD_CLASS] = (cp ? cp->name : Q);
    a = get_arena (o->arena_id);
    if (a) {
	text[OD_ARENA] = a->arena_name;
    } else {
	snprintf (sarena, 64, "%d", o->arena_id);
	text[OD_ARENA] = sarena;
    }
    snprintf (is_array, 64, "%d", (o->is_array ? 1 : 0));
    text[2] = is_array;
    snprintf (size, 64, "%d", o->size);
    text[3] = size;
    text[4] = (c ? c->name : Q);
    text[5] = (m ? m->jmpname : SYS_INIT);
    text[OD_VARIABLE] = variable;
    
    gtk_tree_store_append (list, iter, parent);
    gtk_tree_store_set (list, iter,
			OD_CLASS, text[OD_CLASS],
			OD_ARENA, text[1],
			OD_IS_ARRAY, text[2],
			OD_SIZE, text[3],
			OD_ALLOC_CLASS, text[4],
			OD_ALLOC_METHOD, text[5],
			OD_VARIABLE, text[OD_VARIABLE],
			OD_OBJECT, o,
			-1);
    
}

static void append_string (GtkTreeStore* list, GtkTreeIter* iter, GtkTreeIter* parent, 
			   unsigned char* txt, int len) {
    /* titles = { "Class name", "arena", "is array", "size", "allocating class", "allocating method"};*/    
    glong items_read;
    glong items_written;
    gchar *text[ODN_COLUMNS];
    GError* err = NULL;

    /* I'm not sure if it really is utf16, but that seems to work quite ok */
    gchar* converted_text = 
	g_utf16_to_utf8 ((gunichar2*)txt, len, &items_read, 
			 &items_written, &err);
    if (err != NULL) {
	fprintf (stderr, "Unable to convert text: %s\n", err->message);	
	g_error_free (err);
    }
    text[OD_CLASS] = converted_text;
    text[OD_ARENA] = "";
    text[OD_IS_ARRAY] = "";
    text[OD_SIZE] = "";
    text[OD_ALLOC_CLASS] = "";
    text[OD_ALLOC_METHOD] = "";
    text[OD_VARIABLE] = "";

    gtk_tree_store_append (list, iter, parent);
    gtk_tree_store_set (list, iter,
			OD_CLASS, text[OD_CLASS],
			OD_ARENA, text[1],
			OD_IS_ARRAY, text[2],
			OD_SIZE, text[3],
			OD_ALLOC_CLASS, text[4],
			OD_ALLOC_METHOD, text[5],
			OD_VARIABLE, text[OD_VARIABLE],
			OD_OBJECT, NULL,
			-1);
    g_free (converted_text);
}

static void append_object_dump (obj* o, wret* w) {
    GtkTreeIter   iter;
    append_object (w->clist, &iter, NULL, o, "");
}

static void print_if_method (obj* o, wret* w) {
    if (o->method == (method*)w->data)
	append_object_dump (o, w);
}

static void print_if_class (obj* o, wret* w) {
    if (obj_get_class_id (o) == ((cls*)w->data)->class_id)
	append_object_dump (o, w);
}

static void add_column (GtkWidget* tree, char* label, int pos, 
			gpointer data, void (*cb)(GtkWidget*, gpointer), 
			gint width) {
    GtkCellRenderer *renderer = gtk_cell_renderer_text_new ();
    GtkTreeViewColumn *column = 
	gtk_tree_view_column_new_with_attributes (label, renderer,
						  "text", pos, 
						  NULL );
    gtk_tree_view_append_column (GTK_TREE_VIEW (tree), column);
    gtk_tree_view_column_set_resizable (column, TRUE);
    gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_FIXED);
    gtk_tree_view_column_set_min_width (column, width);
    if (cb != NULL)
	g_signal_connect (G_OBJECT (column), "clicked", G_CALLBACK (cb), data);
}

static void od_column_clicked (GtkWidget *treeviewcolumn, gpointer user_data) {
    GtkTreeView* tree;
    GtkTreeModel* clist;
    int column;
    tree = GTK_TREE_VIEW (user_data);
    clist = gtk_tree_view_get_model (tree);
    for (column = 0; ; column++) {
	if (treeviewcolumn == (GtkWidget*)gtk_tree_view_get_column (tree, column))
	    break;
    }
    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (clist), column, GTK_SORT_ASCENDING);
    gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (clist), column, obj_comprs[column], NULL, NULL);
}

static void instance_row_changed (GtkTreeSelection *selection, gpointer data) {
    GtkTreeIter iter;
    GtkTreeModel *model;
    
    if (gtk_tree_selection_get_selected (selection, &model, &iter)) 
	gtk_tree_model_get (model, &iter, OD_OBJECT, 
			    &last_selected_object, -1);
    else 
	last_selected_object = NULL;
}

static char* get_field (down_link* dl) {
    if (dl->fld)
	return dl->fld->field_name;
    snprintf (buf, 64, "%d", dl->pos);
    return buf;
}

static void expand_node (GtkTreeStore* treestore, GtkObject* list, GtkTreeIter* node) {
    GtkTreePath* path;
    path = gtk_tree_model_get_path (GTK_TREE_MODEL (treestore), node);
    gtk_tree_view_expand_row (GTK_TREE_VIEW (list), path, FALSE); 
    gtk_tree_path_free (path);			
}

/** request to show all live objects by a certain class. */
static void inspect_instance (GtkObject *list) {
    if (last_selected_object != NULL) {
	GtkTreeIter iter, tmp;
	GtkTreeIter parent;
	GtkTreeModel *model;
	GtkTreeSelection *selection;
	gboolean valid;
	down_link* dl;
	get_instance_info (last_selected_object);
	dl = get_last_down_link ();
	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (list));
	if (gtk_tree_selection_get_selected (selection, &model, &parent)) {
	    GtkTreeStore *treestore = (GtkTreeStore*)gtk_tree_view_get_model (GTK_TREE_VIEW (list));
	    if (gtk_tree_model_iter_has_child (GTK_TREE_MODEL (treestore), &parent)) {
		gtk_tree_model_iter_children (GTK_TREE_MODEL (treestore), &iter, &parent);
		valid = TRUE;
		while (valid) {
		    memcpy (&tmp, &iter, sizeof (GtkTreeIter));
		    valid = gtk_tree_model_iter_next (GTK_TREE_MODEL (treestore), &iter);
		    gtk_tree_store_remove (treestore, &tmp);
		}
	    }
	    while (dl != NULL) {
		/*
		fprintf (stderr, "down_link: %s: %s: ", 
			 dl->fld->field_name, 
			 get_field_type (dl->fld));
		*/
		switch (dl->type) {
		case JVMPI_NORMAL_OBJECT: {
		    obj* o = get_object (dl->value.o);
		    if (o) {
			append_object (treestore, &iter, &parent, o, get_field (dl));
			expand_node (treestore, list, &parent);
		    }
		    break;
		}
		case JVMPI_CLASS: {
		    obj* o = get_object (dl->value.o);
		    if (o) {
			append_object (treestore, &iter, &parent, o, get_field (dl));
			expand_node (treestore, list, &parent);
		    }
		    break;
		}
		case JVMPI_BOOLEAN: 
		    /*fprintf (stderr, "%s", dl->value.u1 ? _("true") : _("false"));*/
		    break;
		case JVMPI_BYTE: 
		    /*fprintf (stderr, "%d", dl->value.u1);*/
		    break;
		case JVMPI_CHAR: 
		    /*fprintf (stderr, "%d (%c)", dl->value.u2, (char)dl->value.u2);*/
		    break;
		case JVMPI_SHORT: 
		    /*fprintf (stderr, "%d", dl->value.u2);*/
		    break;
		case JVMPI_INT: 
		    /*fprintf (stderr, "%d", dl->value.u4);*/
		    break;
		case JVMPI_LONG: 
		    /*fprintf (stderr, "%lld", dl->value.u8);*/
		    break;
		case JVMPI_FLOAT:
		    /*fprintf (stderr, "%f", (float)dl->value.u4);*/
		    break;
		case JVMPI_DOUBLE: 
		    /*fprintf (stderr, "%f", (jdouble)dl->value.u4);*/
		    break;
		case JVMPI_GC_PRIM_ARRAY_DUMP:
		    append_string (treestore, &iter, &parent, dl->value.txt, dl->pos);
		    expand_node (treestore, list, &parent);
		}	    
		/*fprintf (stderr, "\n");*/
		dl = dl->next;
	    }
	}
	free_last_down_link ();
    }
}

static GtkWidget *build_menu (GtkWidget* object_list) {
    GtkWidget *menuitem;
    GtkWidget *omenu = gtk_menu_new ();
    menuitem = gtk_menu_item_new_with_label (_("inspect instance"));
    gtk_menu_append (GTK_MENU (omenu), menuitem);
    gtk_widget_show (menuitem);
    gtk_signal_connect_object (GTK_OBJECT (menuitem), "activate",
			       GTK_SIGNAL_FUNC (inspect_instance),
			       GTK_OBJECT (object_list));    
    return omenu;
}

static gint instance_button_handler (GtkWidget *widget,
				     GdkEventButton *event,
				     gpointer  callback_data) {
    if (event->button == 3 && last_selected_object != NULL) {
	GtkWidget* imenu = build_menu (widget);
	gtk_menu_popup (GTK_MENU (imenu), NULL, NULL, NULL, NULL,
                        event->button, event->time);	
	return TRUE;
    }
    return FALSE;
}

wret create_object_dump_window () {
    GtkWidget *obj_dump;
    GtkWidget *scrolledwindow;
    GtkWidget *tree;
    GtkTreeStore *clist1;
    GtkTreeSelection *select;
    wret ret;

    obj_dump = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (obj_dump), _("Object dump"));
    scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
    gtk_container_add (GTK_CONTAINER (obj_dump), scrolledwindow);
    clist1 = gtk_tree_store_new (ODN_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, 
				 G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, 
				 G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER);
    tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (clist1));
    add_column (tree, _("Class"), OD_CLASS, (gpointer)tree, od_column_clicked, 200);
    add_column (tree, _("Arena"), OD_ARENA, (gpointer)tree, od_column_clicked, 80);
    add_column (tree, _("Is array"), OD_IS_ARRAY, (gpointer)tree, od_column_clicked, 80);
    add_column (tree, _("Size"), OD_SIZE, (gpointer)tree, od_column_clicked, 80);
    add_column (tree, _("Allocating class"), OD_ALLOC_CLASS, (gpointer)tree, od_column_clicked, 200);
    add_column (tree, _("Allocating method"), OD_ALLOC_METHOD, (gpointer)tree, od_column_clicked, 200);
    add_column (tree, _("Variable"), OD_VARIABLE, (gpointer)tree, od_column_clicked, 200);
    gtk_tree_view_set_headers_clickable (GTK_TREE_VIEW (tree), TRUE);
    gtk_container_add (GTK_CONTAINER (scrolledwindow), tree);    
    select = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));
    g_signal_connect (G_OBJECT (select), "changed",
		      G_CALLBACK (instance_row_changed),
		      clist1);
    gtk_signal_connect (GTK_OBJECT(tree), "button_press_event",
			GTK_SIGNAL_FUNC (instance_button_handler), 
			clist1);
    gtk_widget_set_usize (obj_dump, 800, 200);
    ret.window = obj_dump;
    ret.clist = clist1;
    return ret;
}

void show_objects_alloced_by_method (hashtab* objects, method* m) {
    wret w = create_object_dump_window ();
    w.data = m;
    w.row = 0;
    jmphash_lock (objects);
    jmphash_for_each_with_arg ((jmphash_iter_fa)print_if_method, objects, &w);
    jmphash_unlock (objects);
    gtk_widget_show_all (w.window);
}

void show_objects_alloced_by_class (hashtab* objects, cls* c) {
    wret w = create_object_dump_window ();
    w.data = c;
    w.row = 0;
    jmphash_lock (objects);
    jmphash_for_each_with_arg ((jmphash_iter_fa)print_if_class, objects, &w);
    jmphash_unlock (objects);
    gtk_widget_show_all (w.window);
}

/* Emacs Local Variables: */
/* Emacs mode:C */
/* Emacs c-indentation-style:"gnu" */
/* Emacs c-hanging-braces-alist:((brace-list-open)(brace-entry-open)(defun-open after)(substatement-open after)(block-close . c-snug-do-while)(extern-lang-open after)) */
/* Emacs c-cleanup-list:(brace-else-brace brace-elseif-brace space-before-funcall) */
/* Emacs c-basic-offset:4 */
/* Emacs End: */
